//*****************************************************************************
//* =====================
//*  HtoZZSpring
//* =====================
//*  
//* (Description)
//*    H -> ZZ* generator
//*
//* (Update Record)
//*    2024/02/13  K.Fujii	Original version.
//*
//*****************************************************************************

#include "JSFSteer.h"
#include "HtoZZSpring.h"

#include "TRandom.h"

#include <sstream>
#include <iomanip>
//#define __DEBUG__
//#define __ZEROWIDTH__
//#define __PHASESPACE__
#ifdef __PHASESPACE__
#define __NODECAY__
#endif
#ifdef __NODECAY__
#define __ZEROWIDTH__
#endif

//*----------------------------------------------------------------------
//*     Numerical and Natural Constants
//*----------------------------------------------------------------------
#include "GENNumCon.h"

ClassImp(HtoZZSpring)
ClassImp(HtoZZSpringBuf)
ClassImp(HtoZZBases)

//-----------------------------------------------------------------------------
// ==============================
//  class HtoZZSpring
// ==============================
//_____________________________________________________________________________
// --------------------------
//  C-tor
// --------------------------
HtoZZSpring::HtoZZSpring(const char       *name,
                         const char       *title,
                               HtoZZBases *bases)
           : JSFSpring(name, title, bases)
{
  fEventBuf = new HtoZZSpringBuf("HtoZZSpringBuf",
                                 "HtoZZSpring event buffer",
                                 this);
  if (!bases) { 
    SetBases(new HtoZZBases);
  }
}

//_____________________________________________________________________________
// --------------------------
//  D-tor
// --------------------------
HtoZZSpring::~HtoZZSpring()
{
  //delete fEventBuf;   // JSFSpring takes care of deleting these
  //delete GetBases();
}

//_____________________________________________________________________________
// --------------------------
//  Initialize
// --------------------------
Bool_t HtoZZSpring::Initialize()
{
  JSFSpring::Initialize();

  if (fFile->IsWritable()) {
    HtoZZBases *bs = static_cast<HtoZZBases *>(GetBases());
    TDirectory *last = gDirectory;
    fFile->cd("/conf");
    TList *dlist = gDirectory->GetListOfKeys();
    if (!dlist->FindObject("init")) gDirectory->mkdir("init");
    fFile->cd("/conf/init");
    bs->Write();
    last->cd(); 
    cerr << ">>>>>> HtoZZBases written to file" << endl;
  }

  return kTRUE;
}

//-----------------------------------------------------------------------------
// ==============================
//  class HtoZZSpringBuf
// ==============================
//_____________________________________________________________________________
// --------------------------
//  SetPartons
// --------------------------
Bool_t HtoZZSpringBuf::SetPartons()
{
  TClonesArray &partons = *fPartons;
  HtoZZBases   *bases   = (HtoZZBases*)((HtoZZSpring*)Module())->GetBases();

  // ----------------------------------------------
  //  Set 4-momenta of final state particles in CM
  // ----------------------------------------------
  fNparton = 7;

  static ANL4DVector *pv = 0;
  if (!pv) {
    pv = new ANL4DVector [fNparton];
  }
  pv[3] = bases->fP[0];  // f1
  pv[4] = bases->fP[1];  // f2
  pv[5] = bases->fP[2];  // f3
  pv[6] = bases->fP[3];  // f4

  pv[1] = pv[2] + pv[3]; // Z1
  pv[2] = pv[4] + pv[5]; // Z2
  pv[0] = pv[1] + pv[2]; // h

  fCosTheta     = bases->GetCosTheta();
  fPhi          = bases->GetPhi();
  fQ2Z1         = bases->GetQ2Z1();
  fCosThetaZ1F   = bases->GetCosThetaZ1F();
  fPhiZ1F        = bases->GetPhiZ1F();
  fQ2Z2         = bases->GetQ2Z2();
  fCosThetaZ2F   = bases->GetCosThetaZ2F();
  fPhiZ2F        = bases->GetPhiZ2F();

  // ----------------------------------------------
  //  Set final state parton infomation
  // ----------------------------------------------


  static TVector **qp = 0;
  if (!qp) {
    qp = new TVector* [fNparton];
    for (Int_t i=0; i<fNparton; i++) qp[i] = new TVector(4);
  }
  for (Int_t i=0; i<fNparton; i++) {
    TVector &q = *qp[i];
    q(0) = pv[i].E ();
    q(1) = pv[i].Px();
    q(2) = pv[i].Py();
    q(3) = pv[i].Pz();
  }
  Int_t    idh     = 25;                          // PDG code for H
  Int_t    idz     = 23;                          // PDG code for Z
  Int_t    idf1    = bases->f1Ptr->GetPID   ();   // PDG code for f from Z1
  Double_t chrg1   = bases->f1Ptr->GetCharge();   // f charge
  Double_t m1      = bases->f1Ptr->GetMass  ();   // f mass
  Double_t m2      = m1;                          // fbar mass
  Int_t    idf3    = bases->f3Ptr->GetPID   ();   // PDG code for f from Z2
  Double_t chrg3   = bases->f3Ptr->GetCharge();   // f charge
  Double_t m3      = bases->f3Ptr->GetMass  ();   // f mass
  Double_t m4      = m3;                          // fbar mass

  Int_t    hel1    = bases->fHelFinal[0];         // f1 helicity
  Int_t    hel2    = bases->fHelFinal[1];         // f2 helicity
  Int_t    hel3    = bases->fHelFinal[2];         // f3 helicity
  Int_t    hel4    = bases->fHelFinal[3];         // f4 helicity
  Double_t color1  = bases->f1Ptr->GetColor();    // color factor for f from Z1
  Double_t color2  = bases->f3Ptr->GetColor();    // color factor for f from Z2
  Int_t    islev1  = color1 > 1. ? 201 : 0;  	  // shower level for f from Z1
  Int_t    islev2  = color2 > 1. ? 201 : 0;  	  // shower level for f from Z2
  Int_t    icf1    = 2;                           // color flux id
  Int_t    icf2    = 3;                           // color flux id
  Double_t rq2z1   = pv[1].Mag();
  Double_t rq2z2   = pv[2].Mag();

  Double_t mass    = bases->GetMass();
#if 0
//#ifdef __DEBUG__
  cerr << " -------------------------- " << endl;
  cerr << " 1 pid=" << idh << " m=" << mass  << " Q=" << 0.
       << " pv=(" << (*qp[0])(0) << ", "
                  << (*qp[0])(1) << ", "
                  << (*qp[0])(2) << ", "
                  << (*qp[0])(3) << ") " << endl;
  cerr << " 2 pid=" << idz << " m=" << rq2z1 << " Q=" << 0.
       << " pv=(" << (*qp[1])(0) << ", "
                  << (*qp[1])(1) << ", "
                  << (*qp[1])(2) << ", "
                  << (*qp[1])(3) << ") " << endl;
  cerr << " 3 pid=" << idz << " m=" << rqz2 << " Q=" << 0.
       << " pv=(" << (*qp[2])(0) << ", "
                  << (*qp[2])(1) << ", "
                  << (*qp[2])(2) << ", "
                  << (*qp[2])(3) << ") " << endl;
  cerr << " 4 pid=" << idf1<< " m=" << m1 << " Q=" << chrg1
       << " pv=(" << (*qp[3])(0) << ", "
                  << (*qp[3])(1) << ", "
                  << (*qp[3])(2) << ", "
                  << (*qp[3])(3) << ") " << endl;
  cerr << " 5 pid=" << idf1<< " m=" << m2 << " Q=" << -chrg1
       << " pv=(" << (*qp[4])(0) << ", "
                  << (*qp[4])(1) << ", "
                  << (*qp[4])(2) << ", "
                  << (*qp[4])(3) << ") " << endl;
  cerr << " 6 pid=" << idf1<< " m=" << m3 << " Q=" << chrg3
       << " pv=(" << (*qp[5])(0) << ", "
                  << (*qp[5])(1) << ", "
                  << (*qp[5])(2) << ", "
                  << (*qp[5])(3) << ") " << endl;
  cerr << " 4 pid=" << idf1<< " m=" << m4 << " Q=" << -chrg3
       << " pv=(" << (*qp[6])(0) << ", "
                  << (*qp[6])(1) << ", "
                  << (*qp[6])(2) << ", "
                  << (*qp[6])(3) << ") " << endl;

  TVector qh(4), qz1(4), qz2(4);
  qz1 = *qp[3] + *qp[4];
  qz2 = *qp[5] + *qp[6];
  qh  = qz1 + qz2;
  cerr << " pz1=(" << qz1[0] << ", "
                   << qz1[1] << ", "
                   << qz1[2] << ", "
                   << qz1[3] << ") " << endl;
  cerr << " pz2=(" << qz2[0] << ", "
                   << qz2[1] << ", "
                   << qz2[2] << ", "
                   << qz2[3] << ") " << endl;
  cerr << " ph =(" << qh [0] << ", "
                   << qh [1] << ", "
                   << qh [2] << ", "
                   << qh [3] << ") " << endl;
#endif

  //                              No. PID  Mass   Charge   pv    Nd 1st Mom  hel  col  shower
  new (partons[0]) JSFSpringParton(1, idh, mass ,     0., *qp[0], 0, 0,  0,    0,    0,      0);
  new (partons[1]) JSFSpringParton(2, idz, rq2z1,     0., *qp[1], 2, 3,  1,    0,    0,      0);
  new (partons[2]) JSFSpringParton(3, idz, rq2z2,     0., *qp[2], 2, 5,  1,    0,    0,      0);
  new (partons[3]) JSFSpringParton(3, idf1,  m1 ,  chrg1, *qp[3], 0, 0,  2, hel1, icf1, islev1);
  new (partons[4]) JSFSpringParton(4,-idf1,  m2 , -chrg1, *qp[4], 0, 0,  2, hel2, icf1, islev1);
  new (partons[5]) JSFSpringParton(5, idf3,  m3 ,  chrg3, *qp[5], 0, 0,  3, hel3, icf2, islev2);
  new (partons[6]) JSFSpringParton(6,-idf3,  m4 , -chrg3, *qp[6], 0, 0,  3, hel4, icf2, islev2);

  return kTRUE ;
}

//-----------------------------------------------------------------------------
// ==============================
//  class HtoZZBases
// ==============================
//_____________________________________________________________________________
// --------------------------
//  C-tor
// --------------------------
HtoZZBases::HtoZZBases(const char *name, const char *title)
         : JSFBases   (name, title), 
           fMass      ( 120.),
           fLambda    (1000.),
           fA         (   0.),
           fB         (   0.),
           fBtilde    (   0.),
           fZ1ModesLo ( 1),
           fZ1ModesHi (12),
           fZ2ModesLo ( 1),
           fZ2ModesHi (12),
           fZ1BosonPtr( 0),
           fZ2BosonPtr( 0),
           fQ2Z1      (0.),
           fQ2Z2      (0.),
           fZ1ModePtr (0),
           fZ2ModePtr (0),
           f1Ptr      (0),
           f2Ptr      (0),
           f3Ptr      (0),
           f4Ptr      (0),
           fCosTheta  (0.),
           fPhi       (0.),
           fXQ2Z1     (0.),
           fCosThetaZ1F(0.),
           fPhiZ1F     (0.),
           fXQ2Z2     (0.),
           fCosThetaZ2F(0.),
           fPhiZ2F     (0.)
{
  //  Constructor of bases.  Default parameter should be initialized here
  //
  // --------------------------------------------
  //  Get parameters from jsf.conf, if specified
  // --------------------------------------------

  cout << "Init htozzbases " << endl;
  
  using namespace std;
  stringstream ins(gJSF->Env()->GetValue("HtoZZBases.CosthRange","-1.0 1.0"));
  ins >> fXL[0] >> fXU[0];

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HtoZZBases.PhiOverPiRange","0.0 2.0"));
  ins >> fXL[1] >> fXU[1];

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HtoZZBases.CosthZ1FRange","-1.0 1.0"));
  ins >> fXL[2] >> fXU[2];

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HtoZZBases.PhiZ1FOverPiRange","0.0 2.0"));
  ins >> fXL[3] >> fXU[3];

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HtoZZBases.CosthZ2FRange","-1.0 1.0"));
  ins >> fXL[4] >> fXU[4];

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HtoZZBases.PhiZ2FOverPiRange","0.0 2.0"));
  ins >> fXL[5] >> fXU[5];

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HtoZZBases.MassH","120.")); 	 // M_x [GeV]
  ins >> fMass;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HtoZZBases.Lambda","1000.")); 	 // Lambda [GeV]
  ins >> fLambda;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HtoZZBases.A","0.")); 	 // a
  ins >> fA;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HtoZZBases.B","0.")); 	 // b
  ins >> fB;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HtoZZBases.Btilde","0.")); 	 // btilde
  ins >> fBtilde;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HtoZZBases.Z1ModesLo","1"));    // Z1 decay mode lo
  ins >> fZ1ModesLo;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HtoZZBases.Z1ModesHi","12"));   // Z1 decay mode hi
  ins >> fZ1ModesHi;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HtoZZBases.Z2ModesLo","1"));    // Z2 decay mode lo
  ins >> fZ2ModesLo;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("HtoZZBases.Z2ModesHi","12"));   // Z2 decay mode hi
  ins >> fZ2ModesHi;

  // --------------------------------------------
  //  Registor random numbers
  // --------------------------------------------
  //--
  //  Initial state helicity combination
  //--
  DefineVariable(fHelCombFinal, 0., 1., 1, 1);
  DefineVariable(fZ1DecayMode , 0., 1., 0, 1);
  DefineVariable(fZ2DecayMode , 0., 1., 0, 1);
  DefineVariable(fXQ2Z1       , 0., 1., 0, 1);
  DefineVariable(fXQ2Z2       , 0., 1., 0, 1);

  //--
  //  cos(theta) and phi
  //--
  fXL[1] = fXL[1]*TMath::Pi();
  fXU[1] = fXU[1]*TMath::Pi();
  fXL[3] = fXL[3]*TMath::Pi();
  fXU[3] = fXU[3]*TMath::Pi();
  fXL[5] = fXL[5]*TMath::Pi();
  fXU[5] = fXU[5]*TMath::Pi();

  DefineVariable(fCosTheta  , fXL[0], fXU[0], 0, 1);
  DefineVariable(fPhi       , fXL[1], fXU[1], 0, 1);
  DefineVariable(fCosThetaZ1F, fXL[2], fXU[2], 1, 1);
  DefineVariable(fPhiZ1F     , fXL[3], fXU[3], 0, 1);
  DefineVariable(fCosThetaZ2F, fXL[4], fXU[4], 1, 1);
  DefineVariable(fPhiZ2F     , fXL[5], fXU[5], 0, 1);

  // --------------------------------------------
  //  Set Bases integration parameters
  // --------------------------------------------
  SetNoOfSample(20000);

  SetTuneValue (1.5);
  SetIteration1(0.05, 20);
  SetIteration2(0.05, 50);
}
// --------------------------
//  D-tor
// --------------------------
HtoZZBases::~HtoZZBases()
{
  delete fZ1BosonPtr;
  delete fZ2BosonPtr;
}

//_____________________________________________________________________________
// --------------------------
//  Func
// --------------------------
Double_t HtoZZBases::Func()
{
  //  Bases Integrand

  Double_t bsWeight = 1.; // Jacobian factor

  // --------------------------------------------
  //  Select final state
  // --------------------------------------------
  Double_t weight = 1;

  GENDecayMode *fZ1ModePtr = fZ1BosonPtr->PickMode(fZ1DecayMode, weight, fZ1Mode);
  bsWeight *= weight;
  f1Ptr = static_cast<GENPDTEntry *>(fZ1ModePtr->At(0));
  f2Ptr = static_cast<GENPDTEntry *>(fZ1ModePtr->At(1));
  Double_t m1   = f1Ptr->GetMass();
  Double_t m2   = f2Ptr->GetMass();

  GENDecayMode *fZ2ModePtr = fZ2BosonPtr->PickMode(fZ2DecayMode, weight, fZ2Mode);
  bsWeight *= weight;
  f3Ptr = static_cast<GENPDTEntry *>(fZ2ModePtr->At(0));
  f4Ptr = static_cast<GENPDTEntry *>(fZ2ModePtr->At(1));
  Double_t m3   = f3Ptr->GetMass();
  Double_t m4   = f4Ptr->GetMass();

  // --------------------------------------------
  //  Check if reduced Ecm enough for prod.
  // --------------------------------------------
  if (fMass < m1 + m2 + m3 + m4) {
    return 0.;
  }

  // --------------------------------------------
  //  Select helicity combination
  // --------------------------------------------
  SelectHelicities(weight);
  bsWeight *= weight;

  // --------------------------------------------
  //  Decide Q^2 of internal lines
  // --------------------------------------------

  Double_t mh   = fMass;
  Double_t qmin = m1 + m2;
  Double_t qmax = mh - (m3 + m4);
#ifndef __ZEROWIDTH__
  fQ2Z1 = fZ1BosonPtr->GetQ2BW(qmin, qmax, fXQ2Z1, weight);
#else
  fQ2Z1 = TMath::Power(fZ1BosonPtr->GetMass(),2);
  weight = kPi*fZ1BosonPtr->GetMass()*fZ1BosonPtr->GetWidth();
#endif
  bsWeight *= weight;

  qmin = m3 + m4;
  qmax = mh - sqrt(fQ2Z1);
  fQ2Z2 = fZ2BosonPtr->GetQ2BW(qmin, qmax, fXQ2Z2, weight);
  bsWeight *= weight;

  // --------------------------------------------
  //  Calcuate differential cross section
  // --------------------------------------------

  GENBranch z1branch(fQ2Z1, fCosThetaZ1F, fPhiZ1F, m1*m1, m2*m2);
  GENBranch z2branch(fQ2Z2, fCosThetaZ2F, fPhiZ2F, m3*m3, m4*m4);
  GENBranch cmbranch(mh*mh, fCosTheta  , fPhi  , &z1branch, &z2branch);

  Double_t gamma = DGammaDX(cmbranch);

  // --------------------------------------------
  //  Fill plots
  // --------------------------------------------
  Xh_fill( 1, fCosTheta          , (bsWeight*gamma));
  Xh_fill( 2, fPhi               , (bsWeight*gamma));
  Xh_fill( 3, TMath::Sqrt(fQ2Z1) , (bsWeight*gamma));
  Xh_fill( 4, fCosThetaZ1F       , (bsWeight*gamma));
  Xh_fill( 5, fPhiZ1F            , (bsWeight*gamma));
  Xh_fill( 6, TMath::Sqrt(fQ2Z2) , (bsWeight*gamma));
  Xh_fill( 7, fCosThetaZ2F       , (bsWeight*gamma));
  Xh_fill( 8, fPhiZ2F            , (bsWeight*gamma));
  Xh_fill( 9, (Double_t)fJCombF  , (bsWeight*gamma));
  Xh_fill(10, (Double_t)fZ1Mode  , (bsWeight*gamma));
  Xh_fill(11, (Double_t)fZ2Mode  , (bsWeight*gamma));
  Dh_fill(12, TMath::Sqrt(fQ2Z1) , 
              TMath::Sqrt(fQ2Z2) , (bsWeight*gamma));

  return (bsWeight * gamma);
}

//_____________________________________________________________________________
// --------------------------
//  DGammaDX
// --------------------------
Double_t HtoZZBases::DGammaDX(GENBranch &cmbranch)
{
  // --------------------------------------------
  //  Phase space
  // --------------------------------------------
  Double_t rs = TMath::Sqrt(cmbranch.GetQ2());
  ANL4DVector qcm(rs, 0., 0., 0.);
  GENFrame    cmframe;

  Double_t q2z1  = cmbranch.GetM12();
  Double_t q2z2  = cmbranch.GetM22();
  Double_t cosz1 = cmbranch.GetCosTheta();
  Double_t phiz1 = cmbranch.GetPhi     ();
  GENPhase2 phaseCM(qcm, q2z1, q2z2, cmframe, cosz1, phiz1, 0);
  ANL4DVector pz1 = phaseCM.GetFourMomentum(0);
  ANL4DVector pz2 = phaseCM.GetFourMomentum(1);
  Double_t betaz = phaseCM.GetBetaBar();
  if (betaz <= 0.) return 0.;

  GENBranch &z1branch = *cmbranch.GetBranchPtr(0);
  Double_t cosf1 = z1branch.GetCosTheta();
  Double_t phif1 = z1branch.GetPhi     ();
  Double_t m12   = z1branch.GetM12();
  Double_t m22   = z1branch.GetM22();
  GENPhase2 phaseZ1(pz1, m12, m22, cmframe, cosf1, phif1, 1);
  fP[0] = phaseZ1.GetFourMomentum(0);
  fP[1] = phaseZ1.GetFourMomentum(1);
  fM[0] = TMath::Sqrt(m12);
  fM[1] = TMath::Sqrt(m22);
  Double_t betaf1 = phaseZ1.GetBetaBar();
  if (betaf1 <= 0.) return 0.;

  GENBranch &z2branch = *cmbranch.GetBranchPtr(1);
  Double_t cosf2 = z2branch.GetCosTheta();
  Double_t phif2 = z2branch.GetPhi     ();
  Double_t m32   = z2branch.GetM12();
  Double_t m42   = z2branch.GetM22();
  GENPhase2 phaseZ2(pz2, m32, m42, cmframe, cosf2, phif2, 1);
  fP[2] = phaseZ2.GetFourMomentum(0);
  fP[3] = phaseZ2.GetFourMomentum(1);
  fM[2] = TMath::Sqrt(m32);
  fM[3] = TMath::Sqrt(m42);
  Double_t betaf2 = phaseZ2.GetBetaBar();
  if (betaf2 <= 0.) return 0.;

  // --------------------------------------------
  //  Calcuate differential cross section
  // --------------------------------------------
  Double_t s      = rs * rs;
#ifdef __DEBUG__
  cerr << " fP[0] = (" 
       << fP[0].E () << ","
       << fP[0].Px() << ","
       << fP[0].Py() << ","
       << fP[0].Pz() << ")" << endl;
  cerr << " fP[1] = (" 
       << fP[1].E () << ","
       << fP[1].Px() << ","
       << fP[1].Py() << ","
       << fP[1].Pz() << ")" << endl;
  cerr << " fP[2] = (" 
       << fP[2].E () << ","
       << fP[2].Px() << ","
       << fP[2].Py() << ","
       << fP[2].Pz() << ")" << endl;
  cerr << " fP[3] = (" 
       << fP[3].E () << ","
       << fP[3].Px() << ","
       << fP[3].Py() << ","
       << fP[3].Pz() << ")" << endl;
  ANL4DVector qz1 = fP[0] + fP[1];
  cerr << " qz1 = (" 
       << qz1.E () << ","
       << qz1.Px() << ","
       << qz1.Py() << ","
       << qz1.Pz() << ")" << endl;
  ANL4DVector qz2 = fP[2] + fP[3];
  cerr << " qz2 = (" 
       << qz2.E () << ","
       << qz2.Px() << ","
       << qz2.Py() << ","
       << qz2.Pz() << ")" << endl;
  ANL4DVector pcm = qz1 + qz2;
  cerr << " pcm = (" 
       << pcm.E () << ","
       << pcm.Px() << ","
       << pcm.Py() << ","
       << pcm.Pz() << ")" << endl;
#endif

  // -------------------
  //  Amplitude squared
  // -------------------
  Double_t amp2 = AmpSquared(cmbranch);

  // -------------------
  //  Put them together
  // -------------------
  static const Int_t    kNbr  = 3;
  static const Double_t kFact = k2Pi/(TMath::Power(k4Pi,3*kNbr));

  Double_t identp = 1./2;                            // identical particle factor
  Double_t dPhase = kFact * betaz * betaf1 * betaf2; // phase space factor
  Double_t flux   = 1./(2.* rs);                     // beam flux factor
  Double_t spin   = 1.;                              // spin average for higgs

  Double_t gamma  = identp * flux * spin * amp2 * dPhase; // in [GeV]

  return gamma;
}

//_____________________________________________________________________________
// --------------------------
//  AmpSquared()
// --------------------------
Double_t HtoZZBases::AmpSquared(GENBranch &cmbranch)
{
  Double_t  color = f1Ptr->GetColor() * f3Ptr->GetColor();
#if 0
  Complex_t amp   = FullAmplitude();
  Double_t  amp2  = TMath::Power(abs(amp),2) * color;
#else
  Complex_t amp   = FullAmplitude();
#if 1
  Double_t  amp2  = TMath::Power(abs(amp),2) * color;
#else
#if 0
  Double_t  amp2  = 2*real(fAmpSM * conj(fAmpB)) * color; // Cb
#else
  Double_t  amp2  = real(fAmpSM * conj(fAmpSM)) * color; // SM
#endif
#endif
#endif

#ifdef __NODECAY__
  Double_t q2z    = cmbranch.GetM12();
  Double_t rq2z   = TMath::Sqrt(q2z);
  Double_t gamz   = fZ1BosonPtr->GetWidth();
           amp2  *= (2*kM_z*gamz)/(TMath::Power((rq2z - kM_z)*(rq2z + kM_z),2)
			                          + TMath::Power(kM_z*gamz,2));
  amp2 *= k8Pi; // branch factor correction
#endif
  return amp2;
}

//_____________________________________________________________________________
// --------------------------
//  FullAmplitude()
// --------------------------
Complex_t HtoZZBases::FullAmplitude()
{
   Double_t gamz   = fZ1BosonPtr->GetWidth();

   Double_t qf1    = f1Ptr->GetCharge();
   Double_t t3f1   = f1Ptr->GetISpin();
   Double_t glz1   = -kGz*(t3f1 - qf1*kSin2W);
   Double_t grz1   = -kGz*(     - qf1*kSin2W);

   Double_t qf3    = f3Ptr->GetCharge();
   Double_t t3f3   = f3Ptr->GetISpin();
   Double_t glz2   = -kGz*(t3f3 - qf3*kSin2W);
   Double_t grz2   = -kGz*(     - qf3*kSin2W);

   static const Bool_t kIsIncoming = kTRUE;
   static const Bool_t kIsOutgoing = kFALSE;

   ANL4DVector ph = fP[0] + fP[1] + fP[2] + fP[3];
   HELScalar  hs(ph);

   HELFermion f1 (fP[0], fM[0], fHelFinal [0], +1, kIsOutgoing);
   HELFermion f1b(fP[1], fM[1], fHelFinal [1], -1, kIsIncoming);
   HELVector  z1(f1b, f1, glz1, grz1, kM_z, gamz);

   HELFermion f2 (fP[2], fM[2], fHelFinal [2], +1, kIsOutgoing);
   HELFermion f2b(fP[3], fM[3], fHelFinal [3], -1, kIsIncoming);
   HELVector  z2(f2b, f2, glz2, grz2, kM_z, gamz);

   Complex_t amp = AmpHtoZZ(hs, z1, z2);

   return amp;
}

//_____________________________________________________________________________
// --------------------------
//  AmpHtoZZ()
// --------------------------
Complex_t HtoZZBases::AmpHtoZZ(const HELScalar  &hs,
                              const HELVector  &z1,
                              const HELVector  &z2)
{
   Double_t  gamz  = fZ1BosonPtr->GetWidth();

   //--------------------------------------------------
   // Calculate Amplitudes
   //--------------------------------------------------
#ifdef __PHASESPACE__
   //---------------------------
   // Just Phase Space
   //---------------------------
   Complex_t amp = 1;
#else
   //---------------------------
   // Higgs Decay Amplitude
   //---------------------------
   Double_t gzzh   = kGz*kM_z;
#if 0
   Double_t g1     = gzzh + 2 * kM_z * kM_z * (fA/fLambda);
   Double_t g2     = -2 * (fB/fLambda);
   Double_t g3     = -4 * (fBtilde/fLambda);
#else
   Double_t g1     = gzzh *(1 + fA);
   Double_t g2     = fB/fLambda;
   Double_t g3     = fBtilde/fLambda;
#endif

#ifndef ANOM_ZZH
   Complex_t amp = HELVertex(z1, z2, hs, gzzh);
#else
   Complex_t amp = HELVertex(z1, z2, hs, g1, g2, g3);
#if 0
#else
   fAmpSM = HELVertex(z1, z2, hs, gzzh);
   fAmpB  = HELVertex(z1, z2, hs, 0., g2, 0.);
   fAmpBt = HELVertex(z1, z2, hs, 0., 0., g3);
#endif
#endif
#endif /* end __PHASESPACE__ */

   return amp;
}

//_____________________________________________________________________________
// --------------------------
//  Userin
// --------------------------
void HtoZZBases::Userin()
{
  // --------------------------------------------
  //  Initialize Z decay table
  // --------------------------------------------
  if (!fZ1BosonPtr) fZ1BosonPtr = new GENPDTZBoson();
  for (Int_t m=1; m<=fZ1BosonPtr->GetEntries(); m++) {
     GENDecayMode *mp = fZ1BosonPtr->GetMode(m); 
     if (mp && (m<fZ1ModesLo || m>fZ1ModesHi)) {
        mp->Lock();
     }
  }
  fZ1BosonPtr->DebugPrint();

  if (!fZ2BosonPtr) fZ2BosonPtr = new GENPDTZBoson();
  for (Int_t m=1; m<=fZ2BosonPtr->GetEntries(); m++) {
     GENDecayMode *mp = fZ2BosonPtr->GetMode(m); 
     if (mp && (m<fZ2ModesLo || m>fZ2ModesHi)) {
        mp->Lock();
     }
  }
  fZ2BosonPtr->DebugPrint();

  // --------------------------------------------
  //  Define some plots
  // --------------------------------------------
  Xh_init( 1, fXL[0], fXU[0],       50, "Costh"    );
  Xh_init( 2, fXL[1], fXU[1],       50, "Phi"      );
  Xh_init( 3,    70.,   110.,       50, "Mz1"      );
  Xh_init( 4, fXL[2], fXU[2],       50, "CosthZ1F" );
  Xh_init( 5, fXL[3], fXU[3],       50, "PhiZ1F"   );
  Xh_init( 6,    70.,   110.,       50, "Mz2"      );
  Xh_init( 7, fXL[4], fXU[4],       50, "CosthZ2F" );
  Xh_init( 8, fXL[5], fXU[5],       50, "PhiZ2F"   );
  Xh_init( 9,     0.,     2.,        2, "Helot "   );
  Xh_init(10,     0.,    12.,       12, "Z1 mode"  );
  Xh_init(11,     0.,    12.,       12, "Z2 mode"  );
  Dh_init(12,     0.,   130.,       65, 
		  0.,   130.,       65, "(Mz1,Mz2)");
}

//_____________________________________________________________________________
// --------------------------
//  Userout
// --------------------------
void HtoZZBases::Userout()
{
  cout << "End of HtoZZBases----------------------------------- " << endl
       << "Higgs mass           = " << fMass     << " [GeV]     " << endl
       << "lambda               = " << fLambda   << " [GeV]     " << endl
       << "a                    = " << fA                         << endl
       << "b                    = " << fB                         << endl
       << "btilde               = " << fBtilde                    << endl
       << "Gamma(H->ZZ*)        = " << GetEstimate()  << " +/- "
                                    << GetError()     << " [GeV]" << endl
       << "Number of iterations = " << GetNoOfIterate()           << endl;
}

//_____________________________________________________________________________
// --------------------------
//  SelectHelicities
// --------------------------
void HtoZZBases::SelectHelicities(Double_t &weight)
{
   static const Int_t kNf = 4;
   static const Int_t kFHelComb[kNf][4] = {{-1, +1, -1, +1},
                                           {-1, +1, +1, -1},
                                           {+1, -1, -1, +1},
                                           {+1, -1, +1, -1}};
   fJCombF = (Int_t)(fHelCombFinal*kNf);
   fJCombF = TMath::Min(fJCombF, kNf-1);
   fHelFinal  [0] = kFHelComb[fJCombF][0];
   fHelFinal  [1] = kFHelComb[fJCombF][1];
   fHelFinal  [2] = kFHelComb[fJCombF][2];
   fHelFinal  [3] = kFHelComb[fJCombF][3];
   weight = kNf;
}
