#ifndef ZHSPRING_H
#define ZHSPRING_H
//*****************************************************************************
//* =====================
//*  HtoZZSpring
//* =====================
//*  
//* (Description)
//*    H -> ZZ* generator
//*
//* (Update Record)
//*    2024/02/13  K.Fujii	Original version.
//*
//*****************************************************************************

#include "TNamed.h"
#include "JSFModule.h"
#include "JSFBases.h"
#include "JSFSpring.h"
#include "JSFBeamGeneration.h"

#include "HELLib.h"
#include "GENLib.h"

//_______________________________________________________________________
// =====================
//  class HtoZZBases
// =====================
//-----------------------------------------------------------------------
class HtoZZBases : public JSFBases {
friend class HtoZZSpringBuf;
public:
  // --------------------------------------------------------------------
  //  Member Functions
  // --------------------------------------------------------------------
  // ----------------------
  //  C-tor and D-tor
  // ----------------------
  HtoZZBases(const char *name  = "HtoZZBases", 
             const char *title = "ZH Bases");
  virtual ~HtoZZBases();

  // ----------------------
  //  Getters and Setters
  // ----------------------
  Double_t GetMass       ()        const { return fMass;        }
  Double_t GetLambda     ()        const { return fLambda;      }
  Double_t GetA          ()        const { return fA;           }
  Double_t GetB          ()        const { return fB;           }
  Double_t GetBtilde     ()        const { return fBtilde;      }

  Double_t GetCosTheta   ()        const { return fCosTheta;    }
  Double_t GetPhi        ()        const { return fPhi;         }
  Double_t GetQ2Z1       ()        const { return fQ2Z1;        }
  Double_t GetCosThetaZ1F()        const { return fCosThetaZ1F; }
  Double_t GetPhiZ1F     ()        const { return fPhiZ1F;      }
  Double_t GetQ2Z2       ()        const { return fQ2Z2;        }
  Double_t GetCosThetaZ2F()        const { return fCosThetaZ2F; }
  Double_t GetPhiZ2F     ()        const { return fPhiZ2F;      }

  void     SetMass     (Double_t m      ) { fMass      = m;     }
  void     SetLambda   (Double_t l      ) { fLambda    = l;     }
  void     SetA        (Double_t a      ) { fA         = a;     }
  void     SetB        (Double_t b      ) { fB         = b;     }
  void     SetBtilde   (Double_t bt     ) { fBtilde    = bt;    }

  // ----------------------
  //   Base class methods
  // ----------------------
  virtual void     Userin();   // Bases user initialization
  virtual void     Userout();  // Bases user output 

  Double_t Func();     // Bases integration function.

  // ----------------------
  //   Utility methods
  // ----------------------
private:
  void     SelectHelicities(Double_t &weight);

  Double_t  DGammaDX     (GENBranch &cmbranch);
  Double_t  AmpSquared   (GENBranch &cmbranch);
  Complex_t FullAmplitude();
  Complex_t AmpHtoZZ     (const HELScalar  &hs,
                          const HELVector  &z1,
                          const HELVector  &z2);

private:
  // --------------------------------------------------------------------
  //  Data Members
  // --------------------------------------------------------------------
  // ----------------
  //  Job parameters
  // ----------------
  Double_t fMass;           // m_h    : mass  of H
  Double_t fLambda;         // Lambda : scale of anomalous couplings
  Double_t fA;              // a
  Double_t fB;              // b
  Double_t fBtilde;         // b~

  Int_t    fZ1ModesLo;      // Z1 decay mode lo;
  Int_t    fZ1ModesHi;      // Z1 decay mode hi;
  Int_t    fZ2ModesLo;      // Z2 decay mode lo;
  Int_t    fZ2ModesHi;      // Z2 decay mode hi;

  // ----------------
  //  Particle Data
  // ----------------
  GENPDTZBoson *fZ1BosonPtr;       //! PD table entry of "Z1"
  GENPDTZBoson *fZ2BosonPtr;       //! PD table entry of "Z2"

  // ----------------
  //  Event info
  // ----------------
  Double_t       fQ2Z1;           // q^2 of final state Z1
  Double_t       fQ2Z2;           // q^2 of final state Z2
  GENDecayMode  *fZ1ModePtr;      // pointer to Z1 decay mode
  GENDecayMode  *fZ2ModePtr;      // pointer to Z2 decay mode
  GENPDTEntry   *f1Ptr;           // point to 1st Z1 daughter
  GENPDTEntry   *f2Ptr;           // point to 2nd Z1 daughter
  GENPDTEntry   *f3Ptr;           // point to 1st Z2 daughter
  GENPDTEntry   *f4Ptr;           // point to 2nd Z2 daughter
  Int_t          fHelFinal  [4];  // final   state helicities
  Int_t          fJCombF;         // final  helicity combination
  Int_t          fZ1Mode;         // Z1 decay mode
  Int_t          fZ2Mode;         // Z2 decay mode
  ANL4DVector    fP[4];           // [0,1,2,3] = [f1z1, f2z1, f1z2, f2z2]
  Double_t       fM[4];           // [0,1,2,3] = [m1  , m2  , m3  , m4  ]

  // ----------------
  //  Integ. vars
  // ----------------
  Double_t       fHelCombFinal;   // final   state helicity combination
  Double_t       fZ1DecayMode;    // decay mode selector for Z1
  Double_t       fZ2DecayMode;    // decay mode selector for Z2
  Double_t       fCosTheta;       // cos(theta_x) of Z1 in h rest frame
  Double_t       fPhi;            // phi_x        of Z1 in h rest frame
  Double_t       fXQ2Z1;          // q^2 of final state Z1
  Double_t       fCosThetaZ1F;     // cos(theta_f) in Z1   frame
  Double_t       fPhiZ1F;          // phi_f        in Z1   frame
  Double_t       fXQ2Z2;          // q^2 of final state Z2
  Double_t       fCosThetaZ2F;     // cos(theta_f) in Z2   frame
  Double_t       fPhiZ2F;          // phi_f        in Z2   frame

  Double_t       fXU[6];          //! [0,1,2,3,4,5] = (cosz1, phiz1, cosf1, phif1, cosf2, phif2)
  Double_t       fXL[6];          //!

  // --------------------
  //  Paritall amplitudes
  // --------------------
  Complex_t      fAmpSM; //
  Complex_t      fAmpB;  //
  Complex_t      fAmpBt; //

  ClassDef(HtoZZBases, 1) // Bases for H -> ZZ* process
};

class HtoZZSpring;

//_______________________________________________________________________
// =====================
//  class HtoZZSpringBuf
// =====================
//-----------------------------------------------------------------------
class HtoZZSpringBuf : public JSFSpringBuf {
public:
  // --------------------------------------------------------------------
  //  Member Functions
  // --------------------------------------------------------------------
  // ----------------------
  //  C-tor and D-tor
  // ----------------------
  HtoZZSpringBuf(const char *name   = "HtoZZSpringBuf", 
                const char *title  = "ZH Spring test event buffer",
	        HtoZZSpring *spring = 0)
	     : JSFSpringBuf(name,title,(JSFSpring*)spring) {} 
  virtual ~HtoZZSpringBuf() {}

  // ----------------------
  //  Getters and Setters
  // ----------------------
  Double_t GetCosTheta  ()           const { return fCosTheta;   }
  Double_t GetPhi       ()           const { return fPhi;        }
  Double_t GetQ2Z1      ()           const { return fQ2Z1;       }
  Double_t GetCosThetaZ1F()           const { return fCosThetaZ1F; }
  Double_t GetPhiZ1F     ()           const { return fPhiZ1F;      }
  Double_t GetQ2Z2      ()           const { return fQ2Z2;       }
  Double_t GetCosThetaZ2F()           const { return fCosThetaZ2F; }
  Double_t GetPhiZ2F     ()           const { return fPhiZ2F;      }

  // ----------------------
  //   Base class methods
  // ----------------------
  virtual Bool_t SetPartons();

private:
  // --------------------------------------------------------------------
  //  Data Members
  // --------------------------------------------------------------------
  // ----------------
  //  Event info
  // ----------------
  Double_t fCosTheta;       // cos(theta_z1) in cm  frame
  Double_t fPhi;            // phi_z1        in cm  frame
  Double_t fQ2Z1;           // q^2 of final state Z1
  Double_t fCosThetaZ1F;     // cos(theta_f)  in Z1  frame
  Double_t fPhiZ1F;          // phi_f         in Z1  frame
  Double_t fQ2Z2;           // q^2 of final state Z2
  Double_t fCosThetaZ2F;     // cos(theta_f)  in Z2  frame
  Double_t fPhiZ2F;          // phi_f         in Z2  frame

  ClassDef(HtoZZSpringBuf, 1)  // HtoZZSpring event buffer
};

//_______________________________________________________________________
// =====================
//  class HtoZZSpring
// =====================
//-----------------------------------------------------------------------
class HtoZZSpring : public JSFSpring {
  // --------------------------------------------------------------------
  //  Member Functions
  // --------------------------------------------------------------------
  // ----------------------
  //  C-tor and D-tor
  // ----------------------
public:
   HtoZZSpring(const char *name  = "HtoZZSpring", 
              const char *title = "ZH Spring test",
              HtoZZBases  *bases = 0);
   virtual ~HtoZZSpring();

   virtual Bool_t Initialize(); // overload the base class method

   ClassDef(HtoZZSpring, 1)  // HtoZZSpring class
};
#endif
